/***********************************************************************************************************************
 *
 * blueMarine Semantic - open source media workflow
 * Copyright (C) 2008-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluemarine.tidalwave.it
 * SCM: https://kenai.com/hg/bluemarine~semantic-src
 *
 **********************************************************************************************************************/
package it.tidalwave.openrdf.elmo.impl;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Modifier;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.logging.Level;
import java.io.Serializable;
import javax.xml.namespace.QName;
import org.openide.util.Lookup;
import org.openrdf.elmo.ElmoManager;
import org.openrdf.elmo.Entity;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.semantic.Wrapper;
import it.tidalwave.openrdf.elmo.ElmoManagerProxy;
import it.tidalwave.openrdf.elmo.ElmoManagerThreadLocal;
import it.tidalwave.openrdf.elmo.ElmoVolatile;

/*******************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 ******************************************************************************/
public class DelegateInvocationHandler implements InvocationHandler, Serializable
  {
    private static final String CLASS = DelegateInvocationHandler.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);

    private static final ElmoManager em = new ElmoManagerProxy();

    /** The entity we're proxying. */
    @Nonnull
    private Entity entityDelegate;
//    private transient Entity entity; FIXME

    /** The id of the entity. */
    @Nonnull
    private final QName qName;

    @CheckForNull
    private Lookup lookup;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public DelegateInvocationHandler (@Nonnull final Entity entityDelegate)
      {
        assert entityDelegate != null : "entityDelegate cannot be null";
        this.entityDelegate = entityDelegate;
        this.qName = entityDelegate.getQName();
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public Object invoke (final @Nonnull Object proxy,
                          final @Nonnull Method method,
                          final @CheckForNull Object[] args)
      throws Throwable // the proxy might throw anything
      {
        if (logger.isLoggable(Level.FINEST))
          {
            logger.finest("invoke(%s@%x, %s, %s)", // don't call this.toString() here
                          proxy.getClass(),
                          System.identityHashCode(proxy),
                          method.getName(),
                          Arrays.toString(args));
          }

        final int argsCount = (args != null) ? args.length : 0;
        final String methodName = method.getName();

        if ("getQName".equals(methodName) && (argsCount == 0))
          {
            return qName;
          }

        if ("getDelegate".equals(methodName) && (argsCount == 0))
          {
            return entityDelegate;
          }

        final Object[] unwrappedArgs = Wrapper.unwrap(args);
        
        if ("getLookup".equals(methodName) && (argsCount == 0))
          {
            synchronized (this)
              {
                if (lookup == null)
                  {
                    lookup = new WrappedLookup(entityDelegate, (Lookup)method.invoke(entityDelegate, unwrappedArgs));
                    logger.finest(">>>> proxied lookup: " + lookup);
                  }
              }

            return lookup;
          }

        if ("hashCode".equals(methodName) && (argsCount == 0))
          {
            return qName.hashCode();
          }

        if ("toString".equals(methodName) && (argsCount == 0))
          {
            return String.format("Proxied@%x[%s]", System.identityHashCode(this), method.invoke(entityDelegate, unwrappedArgs));
          }
        
        if ("equals".equals(methodName) && (argsCount == 1))
          {
            if (!(args[0] instanceof ProxyMarker))
              {
                logger.warning("You're comparing a wrapped object (%s) with a non-wrapped one (%s)\n",
                               "Probably it's not what you mean to do",
                               System.identityHashCode(this), System.identityHashCode(args[0]));
                return false;
              }
            
            final ProxyMarker other = (ProxyMarker)args[0];
            return this.qName.equals(other.getQName());
          }

        renewEntity();
        sanityCheck(method);
        
        Object result = null;

        try
          {
            if (logger.isLoggable(Level.FINEST))
              {
                logger.finest(">>>> invoking method on delegate @%x", System.identityHashCode(entityDelegate));
              }
            
//            else
              {
                result = method.invoke(entityDelegate, unwrappedArgs);
              }
          }
        catch (InvocationTargetException e)
          {
            final Throwable cause = e.getCause();
            throw (cause != null) ? cause : e;
          }

        if (result == null) // FIME: unneeded, Wrapper.wrap is null-safe - put an assertion?
          {
            return null;
          }

        return Wrapper.wrap(result);
      }

    /*******************************************************************************************************************
     *
     * Check whether the {@link Entity} should be renewed. This must happen when it has been created by an
     * {@link ElmoManager} that is different than the current one. To replace the {@code Entity}, a new one is created
     * by searching for the id; then all non-static, non-synthetic and non-volatile fields are copied and the bound
     * {@link Lookup} is flushed.
     *
     ******************************************************************************************************************/
    @Nonnull
    public synchronized void renewEntity()
      throws IllegalArgumentException, IllegalAccessException
      {
        final ElmoManager entityEm = entityDelegate.getElmoManager();
        final ElmoManager currentEm = ElmoManagerThreadLocal.get();

        if (entityEm != currentEm)
          {
            logger.fine(">>>> entity with unmatching ElmoManager: %s: %s vs %s", entityDelegate, entityEm, currentEm);
            final Entity newEntity = em.find(qName); // FIXME: em == currentEm?
            assert newEntity != null : "em.find() returned null for " + qName;
            
            lookup = null; // flush
            logger.fine(">>>> entity @%x is going to replace @%x", System.identityHashCode(newEntity),
                                                                   System.identityHashCode(entityDelegate));

            final Field[] fields = entityDelegate.getClass().getDeclaredFields(); // FIXME: doens't include inherited fields?
//                    logger.finer(">>>> fields to copy: %s, %s", implementationClass, Arrays.toString(fields));

            for (final Field field : fields)
              {
                // FIXME! $ fields should really be excluded, but this breaks non persistent fields.
                // Indeed we should navigate from the entity to all behaviours and preserve their fields.
//                if (((field.getModifiers() & Modifier.STATIC) == 0) && !field.getName().contains("$") && !field.isSynthetic())
                if (((field.getModifiers() & Modifier.STATIC) == 0)
                    && !field.isSynthetic()
                    && !isVolatile(field)
                    && !Lookup.class.isAssignableFrom(field.getType())) // FIXME: drop and put @ElmoVolatile to Lookup
                  {
                    logger.finest(">>>>>>>> copying field %s", field.getName());
                    field.setAccessible(true); // FIXME: is doPrivileged() needed?
                    field.set(newEntity, field.get(entityDelegate));
                  }
              }

            entityDelegate = newEntity;
          }
      }

    /*******************************************************************************************************************
     *
     * Returns {@code true} if the given field is volatile. All fields with Elmo annotations, or with
     * {@link ElmoVolatile}, are volatile.
     *
     * @param  field  the field
     * @return        {@code true} if the field is volatile
     *
     ******************************************************************************************************************/
    private static boolean isVolatile (final @Nonnull Field field)
      {
        for (final Annotation annotation : field.getAnnotations())
          {
            logger.finest(">>>> annotation %s, %s", annotation, annotation.annotationType().getName());

            if (annotation.annotationType().getName().startsWith("org.openrdf.elmo.annotations.")
                || annotation.annotationType().equals(ElmoVolatile.class))
              {
                return true;
              }
          }

        return false;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void sanityCheck (final @Nonnull Method method)
      {
        if (!method.getDeclaringClass().isAssignableFrom(entityDelegate.getClass()))
          {
            logger.warning("Mismatch: entity class %s, method class: %s, instanceof %s",
                           entityDelegate.getClass(), 
                           method.getDeclaringClass(),
                           Arrays.toString(entityDelegate.getClass().getInterfaces()));
          }
      }
  }
