/***********************************************************************************************************************
 *
 * blueMarine Semantic - open source media workflow
 * Copyright (C) 2008-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluemarine.tidalwave.it
 * SCM: https://kenai.com/hg/bluemarine~semantic-src
 *
 **********************************************************************************************************************/
package it.tidalwave.openrdf.elmo.impl;

import javax.annotation.Nonnull;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import javax.xml.namespace.QName;
import org.openrdf.elmo.ElmoManager;
import org.openrdf.elmo.Entity;
import org.openrdf.concepts.owl.Thing;
import org.openrdf.concepts.skos.core.Concept;
import org.openrdf.concepts.skos.core.ConceptScheme;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.semantic.EntityFactory;
import it.tidalwave.semantic.EntityFactory.EntityInitializer;
import it.tidalwave.semantic.OwlThing;
import it.tidalwave.semantic.SkosConcept;
import it.tidalwave.semantic.SkosConceptScheme;
import it.tidalwave.semantic.Wrapper;
import it.tidalwave.openrdf.elmo.ElmoManagerProxy;

/*******************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 ******************************************************************************/
public class ElmoEntityFactory<T, I extends T> extends EntityFactory<T, I>
  {
    private static final String CLASS = ElmoEntityFactory.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);

    private final static Map<Class<?>, Class<?>> roleMap = new HashMap<Class<?>, Class<?>>();

    @Nonnull
    private final Class<T> interfaceClass;

    @Nonnull
    private final Class<I> implementationClass;

    @Nonnull
    private final Class<?>[] extraClasses;

    private final ElmoManager em = new ElmoManagerProxy();

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
      {
        roleMap.put(OwlThing.class, Thing.class);
        roleMap.put(SkosConcept.class, Concept.class);
        roleMap.put(SkosConceptScheme.class, ConceptScheme.class);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public ElmoEntityFactory (final @Nonnull Class<T> interfaceClass,
                              final @Nonnull Class<I> implementationClass,
                              final @Nonnull Class<?> ... extraClasses)
      {
        this.interfaceClass = interfaceClass;
        this.implementationClass = implementationClass;
        this.extraClasses = extraClasses.clone();

        for (int i = 0; i < this.extraClasses.length; i++)
          {
            final Class<?> mapped = roleMap.get(this.extraClasses[i]);

            if (mapped != null)
              {
                this.extraClasses[i] = mapped;
              }
          }

        Wrapper.register(interfaceClass, implementationClass);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public T create(@Nonnull QName qName)
      {
        return create(qName, EntityFactory.<I>emptyInitializer());
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public T create (final @Nonnull QName qName,
                     final @Nonnull EntityInitializer<I> initializer)
      {
        logger.finer("create(%s, %s)", qName, initializer);
        
        if (logger.isLoggable(Level.FINEST))
          {
            logger.finest(">>>> implementationClass: %s extraClasses: %s", implementationClass, Arrays.toString(extraClasses));
          }

        Object object = em.create(qName, implementationClass, extraClasses);
        //
        // Sanity check.
        //
        if (!interfaceClass.isAssignableFrom(object.getClass()))
          {
            final Class<? extends Object> createdClass = object.getClass();
            logger.severe("Bad entity: created class subclass of: %s interfaces: %s", 
                          createdClass.getSuperclass(),
                          Arrays.toString(createdClass.getInterfaces()));
            throw new ClassCastException(String.format("Created entity is not instanceof %s", interfaceClass));
          }
        
        I entity = (I)object;        
        initializer.initialize(entity);
//        entity = em.merge(entity); // FIXME: needed?

        return Wrapper.wrap(entity);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public T find (final @Nonnull QName qName)
      throws NotFoundException
      {
        logger.finer("find(%s)", new Object[]{qName});
        Entity object = this.em.find(qName);

        if (object == null)
          {
            throw new NotFoundException("Can't find entity with id=" + qName);
          }

        if (!(this.interfaceClass.isAssignableFrom(object.getClass())))
          {
            logger.fine(">>>> **** workaround, as a bogus entity has been returned", new Object[0]);
            throw new NotFoundException("Can't find entity with id=" + qName);
          }

        return (T)Wrapper.wrap(object);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public T findOrCreate (final @Nonnull QName qName)
      {
        return findOrCreate(qName, EntityFactory.<I>emptyInitializer());
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public T findOrCreate (final @Nonnull QName qName, final @Nonnull EntityFactory.EntityInitializer<I> initializer)
      {
        try 
          {
            logger.finer("findOrCreate(%s, %s)", new Object[]{qName, initializer});
            return find(qName);
          }
        catch (NotFoundException e)
          {
            logger.finer(">>>> not found", new Object[0]);
            return create(qName, initializer);
          }    
      }
  }

